/********************************** C94.C *********************************

Copyright NCR Corporation Microelectronics Division 12 October 1988

Author:		Ken Stewart NCR MPD Logic Products Applications Engineer
Purpose:	53C94 test software

This software was written by NCR Microelectronics to develop and test new
products.  It is not a general SCSI driver.  NCR assumes no liability for
its use, misuse or inability to use.

Project files:
		C94.C 		Main body of code
		TARGET.C	Target Mode routines
		C94ISR.C	Interrupt Service Routine
		C94TEST.C	Automatic Test Module
		AT_DMA.C	hardware dependant DMA stuff
		AT_DMA.H
		C94AT.H		hardware addresses, global defines
		C94EV.H		external variables


rev 0.0	  01-01-89	Initial release
rev 0.1   01-15-89	added Init buffer with DECREMENTING pattern
rev 1.0	  01-25-89	changed from C94 Mode 0 (C90 mode) to
			mode 1 (16-bit DMA)
rev 1.1	  02-01-89	added Target Mode
rev 1.2	  02-05-89	added FIFO Test
rev 1.3	  02-07-89	added odd-byte and misaligned boundry test
rev 1.4   02-07-89	changed block size: setup(), fillbuf(), dumpbuf()
rev 1.5   07-20-89	added Transfer Counter Test
rev 1.6   07-28-89	added disconnect ability to Initiator
rev 1.7   11-16-89	added Reserve FIFO Byte and Save Residual Byte to
			Odd-Byte Test
rev 1.8   11-16-90	Convert from Microchannel to AT
rev 1.9   04-10-90      Added Automatic Test, mods to FIFO_Test, x_count_test
rev 2.0   10-03-91      Add FAST modes
rev 2.1   12-19-91	changes to INIT
rev 2.2	  02-10-92	add test for misaligned scatter/gather

****************************************************************************/
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <dos.h>
#include <process.h>
#include <ctype.h>
#include <string.h>
#include "c94at.h"


/***************************** Global Variables **************************/

struct parms
  {
  unsigned long	tid, mid, atn, msg_byte, sync_offset, sync_period,
		blksize, blkcount, timeout, parity, clock, burst,
		detail, FASTclk, FAST_SCSI, REQ_ACK, TCT, ANE,
		ESRB, T8, DMA;
  } parms;

int             	i, j, temp1, temp2, page, pagemax;

unsigned long		seg, off, abs_adrs, iocount;

unsigned int		far *dma_buf, cdb_buf[8], byte_count, intstat,
			status, seq, flag, statbyte, msg, x_count, fifo_flags,
			scsicmd;

unsigned char		far* cdb, far* dmabyte, scsi_error, cmd_error,
			pause_flag, SDTR_out[16], SDTR_in[16],
			synch_string[14], test_error, RCB, group_code, bus_ID,
			ID_msg,	config1_image, config2_image, config3_image,
			config4_image;


char			cmd,temp,looprw,setup_cmd;

/*********************************** main **********************************

inputs:  none
outputs: stdout CRT
calls:   almost everything

Description:  This module prints a menu on the standard output (called CRT
through out this program) and prompts the user to enter a letter from the
menu corresponding to some test to perform.  All of the tests are performed
in other modules.

                                ***************/
void main(void)

{


init ();			/* Init variables, chips, motherboard DMA, */
do
  {
  clrscr();
  scsi_error = false;
  printf("\n\n DMA start address = %06lX",abs_adrs); /* troubleshooting */
  printf("\t\t\tchip revision = %02X",RCB);
  printf("\n\n     ********************* NCR 53C96 AT Board Test Menu ********************\n");
  printf("\n  b .................... display buffer memory");
  printf("\n  c .................... DMA CDB into FIFO");
  printf("\n  d .................... init buffer with decrementing pattern");
  printf("\n  e .................... execute an SCSI command");
  printf("\n  f .................... FIFO Test");
  printf("\n  i .................... init buffer with incrementing pattern");
  printf("\n  j .................... fill buffer with alternating 0C 0D");
  printf("\n  l .................... loop on read or write");
  printf("\n  o .................... odd-byte and misaligned boundary test");
  printf("\n  O .................... misaligned and scattered synchronous read");
  printf("\n  r .................... Read 53C94 Registers");
  printf("\n  R .................... Reset SCSI bus");
  printf("\n  s .................... setup menu");
  printf("\n  t .................... target mode");
  printf("\n  x .................... Transfer Counter Test");
  printf("\n  4 .................... Automatic Test");
  printf("\n  q .................... Quit; exit this program");
  printf("\n  v .................... Wr/Rd 1 to SRB, config3 bit4");
  printf("\n  w .................... Wr/Rd 0 to SRB, config3 bit4");
  printf("\n\n ? ");

  tryagain:
  cmd=getche();

  do					/* clear unexpected interrupts */
    {
    status = inportb(stat_reg);
    if (status & 0x80)
      {
      intstat = inportb(int_reg);
      printf("INTERRUPT ");
      }
    } while (status & 0x80);

  switch (cmd)
    {
      case 'b' :  dumpbuf();		/* display buffer memory */
      break;

      case 'c' :
	{
	do                            		/* Get r or w from the	 */
	  {					/* keyboard.  Repeat the */
	  printf("\nread or write ? ");		/* prompt and last input */
	  looprw=tolower(getche());			/* until get r or w      */
	  } while (looprw != 'r' && looprw != 'w');
	DMA_CDB(looprw);		/* DMA Select without ATN */
	}
      break;

      case 'd' :	/* decrement buffer, start at keyboard input */
	{
	printf("\n Initialize memory buffer with an decrementing pattern");
	printf("\n Enter starting value in hex ? ");
	scanf("%x%c",&j,&temp);	/* temp is to discard carriage return */

	fillbuff(cmd,j);
	}
      break;

      case 'e' :	/* Execute an SCSI command */
	{
	scsi_menu();
	if(cmd_error) printf("\n   Unsupported Command !");
	}
      break;

      case 'f' :
	{
	FIFO_Test();
	pause_flag = true;
	}

      break;

      case 'i' :	/* increment buffer, start at keyboard input */
	{
	printf("\n Initialize memory buffer with an incrementing pattern");
	printf("\n Enter starting value in hex ? ");
	scanf("%x%c",&j,&temp);	/* temp is to discard carriage return */

	fillbuff(cmd,j);
	}
      break;

      case 'j' :		/* init buffer with alternating 0C 0D hex */
      fillbuff(cmd,0);
      break;



      case 'l' :
	{
	fillbuff('i',0);
	looprw = 'r';

	parms.tid = 2;
	load_regs();
	tur();

	parms.tid = 0;
	load_regs();
	tur();
/*
	parms.tid = 2;
	load_regs();
	tur();

	parms.tid = 3;
	load_regs();
	tur();

	parms.tid = 4;
	load_regs();
	tur();
*/
	clrscr();
	printf("Loop Read/Write Test...");
	do
	  {

	    if (parms.tid == 0) parms.tid = 2;
	    else if (parms.tid == 2) parms.tid = 0;
/*	    else if (parms.tid == 2) parms.tid = 3;
	    else if (parms.tid == 3) parms.tid = 0;
	    else if (parms.tid == 4) parms.tid = 0;
*/
	    else parms.tid == 0;

	    if (looprw=='r') looprw = 'w';
	    else if (looprw=='w') looprw = 'r';

	    if (looprw=='r') cdb[0] = 0x08;
	    else             cdb[0] = 0x0a;
	    cdb[1]=0x00;
	    cdb[2]=0x01;
	    cdb[3]=0x00;
	    cdb[4]=parms.blkcount;
	    cdb[5]=0x00;

	    load_regs();
	    DATA_DRIVER(looprw, 6, byte_count);
	  }
	while(!kbhit() && !scsi_error);
	}
      break;



      case 'o' :
	{
	printf("\nOdd-byte and Misaligned Boundry test\n");
	do                            		/* Get r or w from the	 */
	  { 					/* keyboard.  Repeat the */
	  printf("\nread or write ? ");		/* prompt and last input */
	  looprw=tolower(getche());			/* until get r or w      */
	  } while (looprw != 'r' && looprw != 'w');

	if (looprw == 'r') cdb[0]=0x08;		/* SCSI Read Command 	*/
	else cdb[0] =0x0a;
	cdb[1]=0x00;				/* lun = 0, msbaddr = 0	*/
	cdb[2]=0x01;				/* msb address = 01	*/
	cdb[3]=0x00;				/* lsb address = 00	*/
	cdb[4]=parms.blkcount;			/* number blocks to xfer */
	cdb[5]=0x00;				/* reserved byte 	*/

	pause_flag = true;
	odd_byte(looprw, 6);
	}
      break;

      case 'O' :
	{
	printf("\nOdd-byte and Misaligned Boundry during scatter\n");
	esrb();
	}
      break;

      case 'q' : break;	/* need a place holder to avoid default */

      case 'r' : read_regs();		/* read 53C406 registers */
      break;

      case 'R' :			/* SCSI Reset */
	{
	outportb(cmd_reg, scsi_reset);
	init();
	}
      break;


      case 'v' :
       {
	outportb(config3,0x10);
	printf("\nWrote SRB to 1 = %02x",inportb(config3));
       }
      break;

      case 'w' :
       {
	outportb(config3,0x00);
	printf("\nWrote SRB to 0 = %02x",inportb(config3));
       }
      break;

      case 's' : setup();		/* parameter Set-up menu */
      break;

      case 't' :
	{
	target_mode();
	if (scsi_error)
	  {
	  printf("\n\nhit any key to continue");
	  getch();
	  }
	}
      break;

      case 'T' :
       {
	outportb(config2,0x40);
	printf("\nConfig2 = %02x",inportb(config2));
	printf("\nHigh Byte = %02X",inportb(e_reg));
	outportb(e_reg,0xaa);
	outportb(cmd_reg,nop);
	printf("\nHigh Byte after 0xaa write = %02X",inportb(e_reg));
	outportb(config2,0x00);
	printf("\config2 after 0x00 write = %02x",inportb(config2));
	outportb(e_reg,0x55);
	outportb(cmd_reg,nop);
	printf("\nHigh Byte after 0x55 write = %02X",inportb(e_reg));
	printf("\n\nhit any key to continue");
	getch();

       }


      case 'x' : x_count_test();	/* Transfer Counter Test */
      break;

      case '4' : test_driver();		/* Automatic Test */
      break;

      default:
	{
	printf("\n\nInvalid response.  Enter a case sensitive letter ");
	printf("from the menu.\n\n ? ");
	goto tryagain;
	}
    }					/* end of switch function  */
    if(scsi_error)
      {
      outportb(cmd_reg, scsi_reset);
      outportb(cmd_reg, soft_reset);
      outportb(cmd_reg, nop);
      load_regs();
      }

      if (pause_flag==true)		/* pause to view results on CRT */
	{
	printf("\nhit any key to return to main menu");
	temp = getch();
	pause_flag=false;
	}

  } while(cmd != 'q');			/* end of do-while loop */
}					/* end of main */

/******************************* Initialize *********************************

inputs:  none
outputs: none
calls:   initIBMdma(), load_regs(), init86C01() *microchannel version only*

			       ***************/

void init(void)

{
dma_buf = (unsigned int far) malloc(0xfff);
dmabyte = (unsigned char *) dma_buf;  /* align char array to word boudary */
cdb     = (unsigned char *) cdb_buf;

parms.mid = 6;				/* my ID */
parms.tid = 0;				/* other device ID */

parms.atn = 'y';			/* Select without ATN */
parms.msg_byte =0x80;			/* ID Message Byte = 0 */
parms.blksize = 512;			/* number of bytes in data block */
parms.blkcount = 1;			/* number of blocks to transfer */
parms.timeout = 0x99;			/* 250 milliseconds */
parms.sync_period = 5;			/* transfer period set to minimum */
parms.sync_offset = 0;			/* asynchronous data transfer */
parms.parity = 'n';			/* enable automatic parity checking */
parms.T8 = 'n';				/* Threshold-8 mode */
parms.burst = 'n';			/* DMA Burst Mode */
parms.detail = 'y';			/* Display lots of info during SCSI */
parms.FASTclk = 'n';			/* assume 25 MHz clock */
parms.FAST_SCSI = 'n';			/* assume SCSI-1 transfer rate */
parms.clock = 0;			/* clock conversion factor, 40 MHz */
parms.REQ_ACK = 0;                      /* req_ack delay */
parms.TCT = 'n';			/* Transfer Counter Test */
parms.ANE = 'n';			/* Active Negation Enable */
parms.DMA = 'n';				/* use DMA for data phase */

outportb(cmd_reg, soft_reset);
outportb(cmd_reg,nop);
outportb(config2, 0x40);		/* enable features */
outportb(cmd_reg, 0x80);
RCB = inportb(e_reg);			/* get Revision code */

outportb(config2,0);
switch (RCB)
  {
  case 0x11 :				/* Proof of Concept Part */
    {
    parms.clock = 0;			/* clock conversion factor, 40 MHz */
    parms.FASTclk = 'y';		/* assume fast clock */
    parms.FAST_SCSI = 'y';		/* workaround bug */
    }
  break;
  case  0x91 :				/* production CF96 */
    {
    parms.clock = 0;			/* clock conversion factor, 40 MHz */
    parms.FASTclk = 'y';		/* assume fast clock */
    }
  break;
  case 0xa2 :
    {
    parms.clock = 0;			/* 40 MHz Clock */
    parms.FASTclk = 'y';
    }
  }

fillbuff('n',0);
load_regs();

initIBMdma(dmabyte, byte_count, 'w');
}

/**************** Load Registers in 53C94 SCSI Controler ***************/

void load_regs(void)

{
outportb(clkconv, parms.clock);			/* clock conversion factor */
outportb(srtimout, parms.timeout);		/* Selection Timeout */

byte_count = parms.blksize * parms.blkcount;	/* number bytes to send */
outportb(tc_low, (byte_count & 0xff));
outportb(tc_high, ((byte_count >> 8) & 0xff));

outportb(dest_ID, parms.tid);			/* Target ID */

outportb(syncprd, parms.sync_period);

outportb(syncoff,((parms.REQ_ACK << 4) | parms.sync_offset));

config1_image=0;				/* load config 1 */
config1_image = config1_image | parms.mid;
if(parms.parity == 'y') config1_image = config1_image | 0x10;
outportb(config1, config1_image);

config2_image=0;                		/* load config 2 */
outportb(config2, config2_image);

					/* DEBUG workaround bug in POC */
/*if (RCB == 0x11 && parms.FASTclk == 'y') parms.FAST_SCSI = 'y';*/

config3_image = 0;				/* load config 3 */
if(parms.T8 == 'y')        config3_image = config3_image | 1;
if(parms.burst=='y')       config3_image = config3_image | 3;	/* must have both bits set */
if(parms.FAST_SCSI == 'y') config3_image = config3_image | 0x10;
if(parms.FASTclk   == 'y') config3_image = config3_image | 0x08;
//outportb(config3, config3_image | 0x04);	/* Save Residual Byte */


if (parms.FAST_SCSI == 'y') parms.ANE = 'y';	/* Active Negation */

config4_image = 0;					/* load Config 4 */
if (parms.TCT == 'y') config4_image = config4_image | 0x02;
if (parms.ANE == 'y') config4_image = config4_image | 0x04;
outportb(config4, config4_image);

}

/******************************** fill buffer *******************************
fillbuff(pattern, start)

input arguemnt:	start   = starting value of pattern
		pattern = type of pattern to fill the memory buffer with.

			  i = incrementing pattern starting at keyboard input.
			  d = decrementing pattern starting at keyboard input.
			  j = alternating 0C 0D (hex).


				**************/
void fillbuff(unsigned char pattern, unsigned int start)
{
byte_count = parms.blksize * parms.blkcount;
switch (pattern)
  {
  case 'i' :
    {
    for (i=0; i<byte_count; ++i) dmabyte[i] = start + i;
    }
  break;

  case 'd' :
    {
    for (i=0; i<byte_count; ++i) dmabyte[i] = start - i;
    }
  break;

  case 'n' :
    {
    for (i=0; i<byte_count; ++i) dmabyte[i] = 0x7F;
    }
  break;

  case 'j' :
    {
    for (i=0; i<byte_count; i=i+2) dmabyte[i] = 0x00;
    for (i=1; i<byte_count; i=i+2) dmabyte[i] = 0xFF;
    }
  }	/* end of switch statement */
}	/* end of fillbuf() */

/******************************* dump buffer ********************************

dumpbuf()

This routine prints the memory array out to the CRT.  There are a varible
number of pages, each of which has 16 rows of 16 bytes displayed in hex.

                                ***************/
void dumpbuf(void)
{
pagemax = byte_count >> 8;	/* total-bytes divided by bytes-per-page */
page = 0;
clrscr();
do
  {
  printf("\n\n");
  for (j=0; j<16; ++j)
    {
    printf("\n");
    for (i=0; i<16; ++i) printf("%02X  ",dmabyte[i + (j<<4) + (page<<8)]);
    }		/* end of j loop */
  ++page;
  printf("\n\n   %d pages left.  ", pagemax - page);
  printf("\n space bar to continue or q to quit");
  temp = tolower(getch());
  }  while (temp != 'q' &&  page < pagemax);
}	/* end of dumpbuf */

/*
void fifo_test(void)
{
printf("\n\nTesting FIFO .............................");
test_error = false;

do
 {
 for (j=0;j<16;++j)
 {
  for (i=0;i<16;++i)
   outportb(fifo,(i+(j<<4)));
  for (i=0;i<16;++i)
  {
   temp1=(inportb(flags_reg) & 0x1f);
   temp2=(inportb(fifo) & 0x00ff);
   if (temp1 != (16-i) || (temp2 != (i+(j<<4)))){
      test_error = true;
      if (parms.detail == 'y')
      {
	printf("ERROR - FIFO = %02x h should be %02x h",temp2,i+(j<<4));
	printf("\n      FLAGS = %02x h should be %02x h",temp1,16-i);
      }
    }
   }
  }
  */
/******************************* FIFO Test *********************************/

void FIFO_Test(void)
{

printf("\n\nTesting FIFO ............................. ");
test_error = false;

for(j=0; j<16; ++j)			/* outer loop */
{
  for(i=0; i<16; ++i) 			/* load 16 bytes into FIFO */
    outportb(fifo,(i + (j << 4)));

  for (i=0; i<16; ++i)		/* check them comming back out */
    {
    temp1= (inportb(flags_reg) & 0x1f);
    temp2= (inportb(fifo) & 0x00ff);
    if (temp1 != (16 - i) || (temp2 != (i + (j << 4))))
      {
      test_error = true;
      if (parms.detail == 'y')
	{
        printf("\nERROR  FIFO  = %02x h  should be %02x h",temp2,i+(j<<4));
	printf("\n       flags = %02d d  should be %02d d\n",temp1,16-i);
	}
      }
     }	/* end of i empty fifo loop */
   }		/* end of j loop */
if (test_error) printf("FAIL");
else printf("pass");
}		/* end of fifo_test */

/**************************** Transfer Counter Test ************************/
void x_count_test(void)

{
test_error = false;
pause_flag=true;
printf("\nTesting Transfer Counter ................. ");

for (x_count=1; x_count != 0; ++x_count)
  {
  outportb(tc_low, x_count & 0xff);
  outportb(tc_high, (x_count & 0xff00) >> 8);
  outportb(cmd_reg, 0x80);				/* NOP command */
  temp1 = (inportb(tc_high) << 8) | inportb(tc_low);
  if (temp1 != x_count)
    {
    test_error = true;
    if (parms.detail == 'y')
      {
      printf("\nERROR  Counter value = %04X    expected %04X",temp1,x_count);
      }
    }
  }
if (test_error) printf("FAIL");
else printf("pass");
}

/************************** parameter Set-Up menu **************************/

void setup(void)
{
do
  {
  clrscr();
  printf("\n\n     **************************** Set-Up Menu ****************************\n");
  printf("\n  a - select with ATN ....................    %c",parms.atn);
  printf("\n  A - Active Negation Enable.............     %c",parms.ANE);
  printf("\n  b - block size (bytes per block)........ %4d",parms.blksize);
  printf("\n  B - block count (blocks to transfer).... %4d",parms.blkcount);
  printf("\t  total bytes: %ld",parms.blksize * parms.blkcount);
  printf("\n  c - Check Parity .......................    %c",parms.parity);
  printf("\n  D - DMA Transfers during Data Phase.....    %c",parms.DMA);
  printf("\n  f - FAST SCSI ..........................    %c",parms.FAST_SCSI);
  printf("\n  F - FAST clock .........................    %c",parms.FASTclk);
  printf("\n  i - ID message byte ....................   %02X",parms.msg_byte);
  printf("\n  m - My ID ..............................   %02X",parms.mid);
  printf("\n  o - synchronous offset .................   %2d",parms.sync_offset);
  printf("\n  p - synchronous transfer period ........   %2d",parms.sync_period);
  printf("\n  z - clock conversion factor ............   %2d",parms.clock);
  printf("\n  r - Timeout register value..............  %3d",parms.timeout);
  printf("\n  t - Target ID ..........................    %d",parms.tid);
  printf("\n  v - Detailed Messages to CRT ...........    %c",parms.detail);
  printf("\n  1 - REQ/ACK Assert/Deassert nibble .....    %1d",parms.REQ_ACK);
  printf("\n  8 - DMA Burst mode......................    %c",parms.burst);
  printf("\n  q - Exit this menu");
  printf("\n\n ? ");

  tryagain:
  setup_cmd = getche();
  switch (setup_cmd)
    {
    case 'a' :	/* set flag for select with or without Attention */
     {
     do
       {
       printf("\n Select with Attention  (y/n) ?");
       parms.atn = tolower(getche());
       } while (parms.atn != 'y' && parms.atn != 'n'); /* accept only y or n */
     if(parms.atn == 'y')
       {
       printf("\n ID Message byte in hex ?");	/* need Message Byte */
       scanf("%x%c", &parms.msg_byte, &temp);
       }
      }
    break;

    case 'A' :			/* Active Negation Enable */
     {
     do
       {
       printf("\n Enable Active Negation  (y/n) ?");
       parms.ANE = tolower(getche());
       } while (parms.ANE != 'y' && parms.ANE != 'n'); /* accept only y or n */
     }
    break;

    case 'b' :			/* change Block Size */
      {
      do
	{
	printf("\n New Block Size in decimal (4096 max) ? ");
	scanf("%d%c",&parms.blksize,&temp);
	} while (parms.blksize > 16384);
      }		/* no BREAK statement here to next question */

    case 'B' :			/* change Block Count */
      {
      do
	{
	printf("\n Block Size * Block Count = 16384 bytes maximum");
	printf("\n New Block Count in decimal (%d max) ? ",16384/parms.blksize);
	scanf("%d%c",&parms.blkcount,&temp);
	} while (parms.blkcount * parms.blksize > 66000);
      }
    break;

    case 'c' :
      {
      do
	{
	printf("\n Enable automatic Parity Checking ? ");
	parms.parity = tolower(getche());
	} while (parms.parity != 'y' && parms.parity != 'n');
      }
    break;

    case 'D' :
      {
      do
	{
	printf("\n Use DMA during Data Phase ? ");
	parms.DMA = tolower(getche());
	} while (parms.DMA != 'y' && parms.DMA != 'n');
      }
    break;

    case 'f' :
      {
	do
	  {
	  printf("\nSynchronous Transfer greater than 5.0 MB/s (y or n) ? ");
	  parms.FAST_SCSI = tolower(getche());
	  } while (parms.FAST_SCSI != 'n' && parms.FAST_SCSI != 'y');

       }
    break;

    case 'F' :
      {
	do
	  {
	  printf("\nExternal CLK = 40 MHz (y or n) ? ");
	  parms.FASTclk = tolower(getche());
	  } while (parms.FASTclk != 'n' && parms.FASTclk != 'y');
       if (parms.FASTclk == 'y') parms.clock = 0;
       else                      parms.clock = 5;
       }
    break;

    case 'i' :			/* change ID Message Byte */
      {
      printf("\nnew ID Message Byte in hex ? ");
      scanf("%x%c",&parms.msg_byte,&temp);
      }
    break;

    case 'm' : 			/* change My ID */
      {
      do
	{
	printf("\n New ID of this machine in decimal ? ");
	scanf("%d%c",&parms.mid,&temp);
	} while (parms.mid > 7);
      }
    break;

    case 'o' :			/* change Synchronous Offset */
      {
      do
	{
	printf("\nNew synchronous offset in decimal (0-15) ? ");
        scanf("%d%c",&parms.sync_offset,&temp);
	} while (parms.sync_offset > 15);	/* do until valid number */
      }
    break;

    case 'p' :			/* change Synchonous Period */
      {
      do
	{
	printf("\n New Synchronous Period in decimal (4-35) ? ");
	scanf("%d%c",&parms.sync_period,&temp);
        } while ((parms.sync_period < 4) || (parms.sync_period > 35));
      }
    break;

    case 'z' :               /* change clock conversion */
    {
    do
      {
      printf("\n Enter Clock Conversion Factor (0,2,3,4,5,6,7)");
      scanf("%d%c",&parms.clock,&temp);
      } while ((parms.clock == 1) || (parms.clock > 7));
    }
    break;

    case 'q' : break;		/* need a place holder to avoid default */


    case 'r' : /* Modify value in Selection/Reselection Timeout register */
      {
      printf("\nNew decimal value to be loaded into Timeout register ?");
      scanf("%d%c",&parms.timeout,&temp);
      }
    break;

    case 't' :			/* change Target ID */
      {
      do
	{
        printf("\nnew Target ID in decimal (0-7) ? ");
        scanf("%d%c",&parms.tid,&temp);		/* temp is to discard CRLF */
	} while (parms.tid > 7);
      }
    break;

    case 'v' :
      {
      printf("\n Verbose mode during SCSI (y/n) ? ");
      temp = tolower(getche());
      switch (temp)
	{
	case 'n' :
	parms.detail = temp;
	break;

	case 'y' :
	parms.detail = temp;
	break;

        default: printf("\n\nDetail setting not changed");
	} /* end of get character case statement */
      }	  /* end of change Detail setting case */
      break;

    case '1' :			/* REQ/ACK Assert/Deassert nibble */
      {
      do
	{
	printf("\nnew REQ/ACK Assertion/Deasertion timing nibble (hex) ? ");
	scanf("%d%c",&parms.REQ_ACK,&temp);	/* temp is to discard CRLF */
	} while (parms.REQ_ACK > 0x0f);
      }
    break;

    case '8' :			/* DMA Burst Mode */
      {
      do
	{
	printf("\nEnable DMA Burst Mode (y or n) ? ");
	parms.burst = tolower(getche());
	} while (parms.burst != 'y' && parms.burst != 'n');
      }
    break;

    default:
      {
      printf("Invalid response.  Enter a case sensitive letter ");
      printf("from the menu.\n ? ");
      goto tryagain;
      }
    }				/* end of switch(setup_cmd) */
  } while (setup_cmd != 'q');	/* end of do-while loop */
load_regs();			/* load the parameters into the chip */
}				/* end of setup() */

/******************************** SCSI menu ********************************/

void scsi_menu(void)
{

scsi_error = false;
cmd_error = false;
scsicmd = 0;

if (parms.sync_offset > 0) strcpy(synch_string,"Synchronous ");
else strcpy(synch_string,"Asynchronous");

outportb(cmd_reg,0x80);				/* clear TC zero bit */


clrscr();
printf("\n****************************** SCSI Command Menu *****************************\n");
printf("\n  00 ............................. Test Unit Ready");
printf("\n  03 ............................. Request Sense");
printf("\n  08 ............................. Read");
printf("\n  0a ............................. Write");
printf("\n  28 ............................. Extended Read");
printf("\n  2a ............................. Extended Write");
printf("\n  50 ...Synchronous Negotiation... %s",synch_string);
printf("\n\n ? ");

scanf("%x",&scsicmd);

switch(scsicmd)
  {
  case 0x00 :  tur();     			/* Test Unit Ready */
  break;

  case 0x03 :				/* Request Sense */
    {
    printf("\nRequest Sense  (display buffer memory to see sense data)\n");
    cdb[0]=0x03;			/* op-code		*/
    cdb[1]=0x00;			/* lun = 0		*/
    cdb[2]=0x00;			/* reserved		*/
    cdb[3]=0x00;			/* reserved		*/
    cdb[4]=0xff;			/* allocation length	*/
    cdb[5]=0x00;			/* control byte		*/

    rqsense(6);				/* driver for Request Sense */
    }
  break;

  case 0x08 :        			/* SCSI read command */
    {
    cdb[0]=0x08;			/* SCSI Read Command 		*/
    cdb[1]=0x00;			/* lun = 0, msbaddr = 0		*/
    cdb[2]=0x01;			/* msb address = 01		*/
    cdb[3]=0x00;			/* lsb address = 00		*/
    cdb[4]=parms.blkcount;		/* number blocks to transfer	*/
    cdb[5]=0x00;			/* reserved byte		*/
    parms.blkcount = parms.blkcount;
    DATA_DRIVER ('r', 6, byte_count); 	/* driver for data transfer*/
    }
  break;

  case 0x0a :			/* SCSI write command		*/
    {
    cdb[0]=0x0a;			/* opcode */
    cdb[1]=0x00;			/* lun = 0, msbaddr = 0		*/
    cdb[2]=0x01;			/* msb address = 01		*/
    cdb[3]=0x00;			/* lsb address = 00		*/
    cdb[4]=parms.blkcount;		/* number blocks to transfer    */
    cdb[5]=0x00;			/* reserved byte		*/

    DATA_DRIVER ('w', 6, byte_count);	/* driver for data transfer*/
    }
  break;

  case 0x28 : 			/* Extended Read */
    {
    cdb[0]=0x28;			/* Extended Read Command	*/
    cdb[1]=0x00;			/* lun = 0			*/
    cdb[2]=0x00;			/* MSB of block address		*/
    cdb[3]=0x00;			/* 2SB of block address		*/
    cdb[4]=0x01;			/* 3SB of block address		*/
    cdb[5]=0x00;			/* LSB of block address		*/
    cdb[6]=0x00;			/* reserved byte		*/
    cdb[7]=parms.blkcount<<8;		/* MSB of transfer length	*/
    cdb[8]=parms.blkcount;		/* LSB of transfer length	*/
    cdb[9]=0x00;			/* control byte			*/

    DATA_DRIVER ('r', 10, byte_count);	/* driver for data transfer*/
    }
  break;

  case 0x2a : 			/* Extended Write Bench Mark Test */
    {
    cdb[0]=0x2a;			/* Exteded Write Command	*/
    cdb[1]=0x00;			/* lun = 0			*/
    cdb[2]=0x00;			/* MSB of block address		*/
    cdb[3]=0x00;			/* 2SB of block address		*/
    cdb[4]=0x01;			/* 3SB of block address		*/
    cdb[5]=0x00;			/* LSB of block address		*/
    cdb[6]=0x00;			/* reserved byte		*/
    cdb[7]=parms.blkcount<<8;		/* MSB of transfer length	*/
    cdb[8]=parms.blkcount;		/* LSB of transfer length	*/
    cdb[9]=0x00;			/* control byte			*/

    DATA_DRIVER ('w', 10, byte_count); 	/* driver for data transfer*/
    }
  break;

  case 0x50 :					/* Synchronous Negotiation */
    {
    printf("\n\nSynchronous Negotiation");
    if (parms.FASTclk == 'y')
      sync_nego(parms.sync_period * 25, parms.sync_offset);
    else
      sync_nego(parms.sync_period * 40, parms.sync_offset);
    }

  break;

  default :  cmd_error = true;			/* unsupported command */

  }		/* end of switch statement */
}		/* end of scsimenu */

/******************************* Test Unit Ready ************************/
void tur (void)
    {
    pause_flag = true;
    printf("\nTest Unit Ready Command\n");
    cdb[0]=0x00;			/* op-code			*/
    cdb[1]=0x00;			/* lun = 0			*/
    cdb[2]=0x00;			/* reserved			*/
    cdb[3]=0x00;			/* reserved			*/
    cdb[4]=0x00;			/* reserved			*/
    cdb[5]=0x00;			/* control byte			*/

    select(6);				/* select a device */
    isr(5);				/* expect Status In phase */
    if(scsi_error) return;		/* exit, abort */

    cmdcomp();				/* command complete sequence */
    isr(3);				/* expect Status and Message good */
    if(scsi_error) return;		/* exit, abort */

    message();				/* issue message accepted cmd */
    isr(4);				/* expect Disconnected Interrupt */
    }


/********************************* DATA_DRIVER ******************************
DATA_DRIVER (rw, cdb_len, data_len)

input argument rw = phase target is expected to request. r for Data In phase,
                                                         w for Data Out.

	  cdb_len = number of bytes in Command Descripter Block
	 data_len = number of bytes in data block

This is the driver that calls the four routines that program the 53C406 to:
Select a device, Transfer data across the SCSI bus, receive Status and
Message bytes from the Target, and disconnect from the SCSI bus.

			       ***************/
void DATA_DRIVER
(char rw, unsigned short cdb_len, int data_len)
{
pause_flag=true;

select(cdb_len);			/* select a device */
if (rw == 'w') isr(0);			/* expect Data Out Phase */
else           isr(1);			/* expect Data In */
if(scsi_error) return;			/* exit, abort */
while ((status & 7) == 7)		/* Message In ? */
  {
  Message_In(rw);
  if (scsi_error) return;
  }

if (parms.DMA == 'y' || rw == 'r')	/* use DMA to transfer data ? */
  {
  dma_xfer_data(&dmabyte[0], data_len, rw);
  isr(2);
  }
else					/* no, non-DMA trasfer data */
  {
  nonDMA_xfer_data(dmabyte, data_len, rw);
  }
if (scsi_error) return;

cmdcomp();				/* command complete sequence */
isr(3);					/* expect Status and Message good */
if(scsi_error) return;			/* exit, abort */

message();				/* issue message accepted cmd */
isr(4);					/* expect Disconnected Interrupt */
}
/********************************** select *********************************/

void select(unsigned short cdb_len) 	/* Select an SCSI device */
{
scsi_error = false;
if(parms.atn == 'y')
  {
  if(parms.detail == 'y') printf("\nSelection with ATN ... ");
  outportb(cmd_reg, clr_fifo);
  outportb(fifo, parms.msg_byte);		/* load ID Message Byte */
  for (i=0; i < cdb_len; ++i)			/* load CDB into FIFO	*/
    outportb(fifo,cdb[i]);
  outportb(cmd_reg, 0x42);			/* non-DMA Select with ATN */
  }
else
  {
  if(parms.detail == 'y') printf("\nSelection without ATN ... ");
  outportb(cmd_reg, clr_fifo);
  for (i=0; i < cdb_len; ++i) 			/* load CDB into FIFO */
    outportb(fifo,cdb[i]);
  outportb(cmd_reg, 0x41);			/* non-DMA select without ATN */
  }
}	/* end of subroutine select */

/*************************** DMA Transfer Data ****************************
dma_xfer_data (start_addrs, count, rw)

Input Arguments:
	start_addrs = starting addres in memory of data block
	count 	    = number of bytes to transfer
	rw 	    = direction of data flow.  r for Data In, w for Data Out

This routine will transfer the specified number of bytes across the SCSI
bus.  External DMA hardware is expected to transfer data beteween the
C406 FIFO and main memory.

                                ***************/

void dma_xfer_data(char far* start_addrs, int count, char rw)
{
if(parms.detail == 'y')
  {
  if (rw == 'w') printf("\nData Out ... ");
  else printf("\nData In ... ");
  }

initIBMdma(start_addrs, (count), rw);	/* address, byte count, direction */

outportb(tc_low,((count) & 0xff));		/* transfer count lsb	*/
outportb(tc_high,((count) >> 8) & 0xff);		/* transfer count msb	*/
outportb(cmd_reg,0x90);				/* DMA Transfer Info	*/
}

/************************* non-DMA Transfer Data *************************/
void nonDMA_xfer_data(char far* start_addrs, int count, char rw)
{
unsigned int	out_count, fifo_count;

out_count = 0;

if (rw == 'w')
  {						/* Data Out */
  if (parms.detail == 'y') printf("\nnon-DMA Data Out ... ");
  do
    {
    if (count - out_count > 16)		/* more than 16 bytes to send */
      {
      for (i=0; i<16; ++i)			/* fill FIFO */
	{
	outportb(fifo, start_addrs[out_count]);
	++out_count;
	}
      outportb (cmd_reg, Transfer_Info);
      isr(20);
      if (scsi_error) return;
      }
    else				/* 16 or fewer bytes to send */
      {
      fifo_count = count - out_count;
      for (i=0; i<fifo_count; ++i)		/* fill FIFO */
	{
	outportb(fifo, start_addrs[out_count]);
	++out_count;
	}

      outportb (cmd_reg, Transfer_Info);
      if (out_count == byte_count) isr(21);	/* expect status phase */
      else isr(7);    		/* expect data out for odd-byte test */
      if (scsi_error) return;
      }
    } while (out_count < count);
  }

else						/* Data In */
  {
  printf("\nERROR! non-DMA Data In not supported !");
  scsi_error = true;
  }
}
/******************************** cmdcomp **********************************/

void cmdcomp(void)	/* issue the command complete sequence command	*/
{
if(parms.detail == 'y') printf("\nStatus and Message ... ");
outportb(cmd_reg, clr_fifo);
outportb(cmd_reg,0x11);
}

/********************************* message *********************************/

void message(void)		/* issue the message accepted command	*/
{
outportb(cmd_reg, clr_fifo);
outportb(cmd_reg,Clear_ACK);
}


/**************************** odd_byte ***********************************
odd_byte (rw, cdb_len)

inputs:	cdb_len = length of Command Descriptor Block
	     rw = data direction; either r (read) or w

This driver handles both the "odd-byte" and the  "misaligned boundry"
conditions for Initiator mode.  This routine tests the chip (my first
priority) and illustrates the programming technique for odd-byte and
misaligned boundary cases.  This is required only in hardware that does not
support byte control.  For Asynchronous Data In/Out as well as
Synchronous Data Out, It uses Programmed I/O (PIO) to place the first and
last bytes into the FIFO but uses DMA for all the rest.  A more efficient
technique for Asynchronous Data out is to place the first byte into the FIFO
then issue a DMA Transfer Info (in this routine I use a non-DMA transfer info
followed by DMA transfer Info).  This would require one less
interrupt.  For Synchronous Data In, the Reserve FIFO byte option must be
used.  See the manual description of Config-2 bit 7.  The sync Data In case
also tests the Save Residual Byte option.  This option may also be used for
async data transfer.  Neither Reserve FIFO Byte nor Save Residual Byte
existed in silicon prior to

			609-3400486 (53C94)
			609-3400488 (53C95)
			609-3400489 (53C95)



                               ***************/

void odd_byte(char rw, int cdb_len)
{

pause_flag=true;

if(parms.sync_offset > 0 && rw == 'r')         /* if Sync Data In   */
  {
  outportb(config2, config2_image | 0x80);	/* Reserve FIFO byte */
  outportb(config3, config3_image	 | 0x04);	/* Save Residual Byte */
  if(parms.detail == 'y') printf("\n\nspecial Synchronous case");
  }

select(cdb_len);			/* select a device */
if (rw == 'w')  isr(0);			/* expect Data Out Phase */
else            isr(1);			/* expect Data In */
if(scsi_error) return;			/* exit, abort */
while ((status & 7) == 7)			/* Message In ? */
  {
  Message_In(rw);
  if (scsi_error) return;
  }

if(rw == 'w')					/* Data Out */
  {
  if(parms.detail == 'y') printf("\nData Out ... ");

  outportb(fifo,dmabyte[1]);			/* PIO first data byte */
  outportb(cmd_reg,Transfer_Info);
  isr(7);
  if(scsi_error) return;
  if(parms.detail == 'y') printf("complete, first byte");

  if (parms.DMA == 'y')				/* DMA n-2 bytes */
    {
    dma_xfer_data(&dmabyte[2], byte_count-2, rw);
    isr(7);
    }
  else					/* forced non-DMA trasfer data */
    {
    nonDMA_xfer_data(&dmabyte[2], byte_count-2, rw);
    }
    if(scsi_error) return;
  if(parms.detail == 'y') printf("complete, DMA %d bytes",byte_count-2);

  if(parms.detail == 'y') printf("\nData Out ... ");			  	/* PIO the last byte */

  outportb(tc_low, 1);				/* clear TC zero bit */
  outportb(cmd_reg, 0x80);

  outportb(fifo,dmabyte[byte_count]);
  outportb(cmd_reg,Transfer_Info);
  isr(21);
  if(scsi_error) return;
  if(parms.detail == 'y') printf(", last byte");
  }

else							/* Data In */
  {
  if(parms.sync_offset == 0)				/* Asynchronous */
    {
    if(parms.detail == 'y') printf("\nData In ... ");
    outportb(cmd_reg, clr_fifo);			/* PIO first byte */
    outportb(cmd_reg,Transfer_Info);
    isr(7);
    if(scsi_error) return;
    dmabyte[1] = inportb(fifo);
    if(parms.detail == 'y') printf("complete, first byte");

    dma_xfer_data(&dmabyte[2], byte_count-2,rw);	/* DMA n-2 bytes */
    isr(7);
    if(scsi_error) return;
    if(parms.detail == 'y') printf("complete, DMA %d bytes",byte_count-2);

    if(parms.detail == 'y') printf("\nData In ... ");				/* PIO last byte */
    outportb(cmd_reg, clr_fifo);
    outportb(cmd_reg,Transfer_Info);
    isr(2);
    if(scsi_error) return;
    dmabyte[byte_count] = inportb(fifo);
    if(parms.detail == 'y') printf(", last byte");
    }

  else				/* Synchronous Data In */
    {
      /* The low byte at offset=0 in the memory array dmabyte[] is not part of
	the SCSI transfer.  Copy it from memory to register F, the
	 FIFO bottom register.  16-bit DMA will then overwrite the low byte
	 with a copy of itself.	 Since the blocksize is even, there will be an
	 extra, residual, byte after DMA finishes.  After the interrupt,
	 isr(13), the last byte will be in the FIFO.  If the block size is set
	 to max, then the last byte could be written outside the dmabyte[]
	 array, corrupting some other program variable.  Therefore its value
	 is written to the CRT but it is not placed in memory.  */

    outportb(FIFO_bottom,dmabyte[0]);

    dma_xfer_data(&dmabyte[0], byte_count+1, rw);
    isr(13);
    if(scsi_error) return;
    printf("\nData In ... complete, last byte = %02X",inportb(fifo));
    outportb(cmd_reg,0x80);		/* clear Transfer Count Zero bit */

    }
  }

cmdcomp();				/* command complete sequence */
isr(3);				/* expect Status and Message good */
if(scsi_error) return;		/* exit, abort */

message();				/* issue message accepted cmd */
isr(4);				/* expect Disconnected Interrupt */
}

/****************************** DMA CDB into FIFO ****************************
DMA_CDB(rw)

inputs	rw = direction indicator: r for SCSI read or w for write
outputs	none

This subroutine loads the SCSI Command Descriptor Block (CDB) into the C406
FIFO using external DMA

                                  ***************/

void DMA_CDB(char rw)
{
  pause_flag=true;

  if(parms.atn == 'y')
    {
    if (parms.detail == 'y') printf("\n\nDMA Select with ATN ... ");

    cdb[0] = parms.msg_byte;
    if (rw=='r') cdb[1] = 0x08;			/* SCSI read or write in  */
    else         cdb[1] = 0x0a;			/* the Command Descriptor */
    cdb[2]=0x12;				/* lun =  , msbaddr =  	  */
    cdb[3]=0x34;				/* msb address = 	  */
    cdb[4]=0x56;				/* lsb address = 	  */
    cdb[5]=parms.blkcount;			/* number of blocks to send*/
    cdb[6]=0x9a;				/* reserved byte	  */

    initIBMdma(cdb,7,'w');
    outportb(tc_low,7);
    outportb(tc_high,0);
    outportb(cmd_reg,0xc2);			/* DMA Select with ATN */
    }

  else
    {
    if (parms.detail == 'y') printf("\n\nDMA Select without ATN ... ");

    if (rw=='r') cdb[0] = 0x08;			/* SCSI read or write in  */
    else         cdb[0] = 0x0a;			/* the Command Descriptor */
    cdb[1]=0x12;				/* lun =  , msbaddr =  	  */
    cdb[2]=0x34;				/* msb address = 	  */
    cdb[3]=0x56;				/* lsb address = 	  */
    cdb[4]=parms.blkcount;			/* number of blocks to send*/
    cdb[5]=0x9a;				/* reserved byte	  */

    initIBMdma(cdb,6,'w');
    outportb(tc_low,6);
    outportb(tc_high,0);
    outportb(cmd_reg,0xc1);			/* DMA Select without ATN */
    }

  if (rw == 'w') isr(14);			/* expect Data Out Phase */
  else           isr(15);			/* expect Data In */
  if(scsi_error) return;			/* exit, abort */

  while ((status & 7) == 7)			/* Message In ? */
    {
    outportb(cmd_reg, 0x80);			/* clear TC-zero bit */
    outportb(cmd_reg, clr_fifo);
    Message_In(rw);
    if (scsi_error) return;
    }

  if (parms.DMA == 'y' || rw == 'r')	/* use DMA to transfer data ? */
    {
    dma_xfer_data(&dmabyte[0], byte_count, rw);
    isr(2);
    }
  else					/* no, non-DMA trasfer data */
    {
    nonDMA_xfer_data(dmabyte, byte_count, rw);
    }
  if (scsi_error) return;

  cmdcomp();				/* command complete sequence */
  isr(3);				/* expect Status and Message good */
  if(scsi_error) return;		/* exit, abort */

  message();				/* issue message accepted cmd */
  isr(4);				/* expect Disconnected Interrupt */
}

/********************************* rqsense **********************************
rqsense(cdb_len)

input argurment cdb_len = length of the command descripter block

This routine is a driver for the Request Sense SCSI command

			       ***************/
void rqsense(unsigned short cdb_len)
{

  char test;

  pause_flag = true;
  select(cdb_len);			/* select a device */
  isr(1);				/* expect Data In */
  if(scsi_error) return;		/* exit, abort */

  dma_xfer_data(&dmabyte[0], 8, 'r');	/* DMA Data In, 8 bytes */
  isr(6);				/* possibly more than 8 bytes */
  if(scsi_error) return;		/* exit, abort */

  cmdcomp();				/* command complete sequence */
  isr(3);				/* expect Status and Message good */
  if(scsi_error) return;		/* exit, abort */

  message();				/* issue message accepted cmd */
  isr(4);				/* expect Disconnected Interrupt */
}

/*********************** Message In after Command Phase *********************
			    DisConnect then ReSelect

Message_In(rw)

input argument:	rw = data direction, r (Read) for Data In, w for Data Out

				 *************/
void Message_In(char rw)
{
if(parms.detail == 'y') printf("\nMessage In ... ");
outportb(cmd_reg,Transfer_Info);
isr(8);
if(scsi_error) return;
switch (temp1)			/* temp1 = message-byte loaded in isr(8) */
  {
  case 0x02 :
    {
    if (parms.detail == 'y')
      {
      printf(" ... Save Data Pointers");
      printf("\nclear ACK ...");
      }
    outportb(cmd_reg,Clear_ACK);
    isr(9);
    }
  break;

  case 0x04 :
    {
    if (parms.detail == 'y')  printf(" ... Disconnect Message");
    outportb(cmd_reg,Clear_ACK);
    isr(4);				/* expect disconnected interrupt */
    if(scsi_error) return;

    outportb(cmd_reg,0x44);		/* enable ReSelection */
    isr(10);				/* wait for ReSelect */
    if(rw == 'r') isr(11);		/* expect Data In interrupt */
    else          isr(12);		/* expect Data Out interrupt */
    }
  break;

  default :
    {
    printf(" ... received unsupported message = %02X hex",temp1);
    scsi_error = true;
    return;
    }
  }	/* end of switch() */
}	/* end of Message_In() */

/************************* Synchronous Negotiation ***********************

sync_nego(period, offset)

inputs:		period = "m" in SCSI spec
		offset = "n" in SCSI spec
outputs:	none

First send a Synchronous Data Transfer Request (SDTR) message to Target.
Then receive Targets SDTR message.  A more general SCSI driver will also
expect a Message Accepted or Message Rejected message in response to our
SDTR message out.  This routine then sets the 53C94 Synchronous Offset
register to value received by Target (called "n" in SCSI spec).  It also
sets C94 Synchronous Transfer Period Register using this formula:

		  STPR = 4m divided by CLK period

In this circuit, CLK is either 25 MHz (40 ns) or 40 MHz (25ns).  So for
25 MHz, STPR = 4m/40 = m/10
40 MHz, STPR = 4m/25
                                  **********/

void sync_nego(unsigned int period, unsigned int offset)
{
pause_flag = true;
SDTR_out[0] = 1;				/* Extended Message message */
SDTR_out[1] = 3;				/* message length */
SDTR_out[2] = 1;				/* SDTR message */
SDTR_out[3] = period / 4;			/* Synchronous Period */
SDTR_out[4] = offset;				/* offset */

					/* Select with ATN and stop */
if(parms.detail=='y') printf("\n\nSelect with ATN and stop ... ");
outportb(cmd_reg,clr_fifo);
outportb(fifo,parms.msg_byte);
for (i=0; i<5; ++i) outportb(fifo,SDTR_out[i]);
outportb(cmd_reg,0x43);
isr(16);
if (scsi_error) return;

			/* Send Synchronous Data Transfer Request message */
if(parms.detail=='y') printf("\nSDTR message ... ");
outportb(cmd_reg,Transfer_Info);
isr(17);
if (scsi_error) return;

						/* get Target's response */
if(parms.detail=='y') printf("\nMessage in ... ");
i = 0;
outportb(cmd_reg,clr_fifo);
do
  {
  outportb(cmd_reg,Transfer_Info);
  isr(18);
  outportb(cmd_reg,Clear_ACK);
  while (!kbhit() && (inportb(stat_reg) < 0x80)); /* Poll for interrupt  */
  intstat = inportb(int_reg);			  /* clear INT 		 */
  i = ++i;					  /* don't care what INT */
  } while (!scsi_error && (inportb(stat_reg) & 7) == 7);  /* Message In? */

fifo_flags = inportb(flags_reg) & 0x1f;
for(i=0; i<fifo_flags; ++i)			/* target response in FIFO */
  {
  SDTR_in[i] = inportb(fifo);
  if(parms.detail=='y') printf("%02Xh ",SDTR_in[i]);
  }
if (scsi_error) return;


/* This test routine expects the Target to return an SDTR message.  For    */
/* general SCSI, the routine should also expect a Message Accepted message */
/* or a Message Reject message.                                            */


if(SDTR_in[0]==1 && SDTR_in[1]==3 && SDTR_in[2]==1)
  {						/* valid SDTR message */
  if(parms.detail=='y')
    {
    printf ("\n  SDTR message received ");
    for (i=1; i<5; ++i) printf(" %02X",SDTR_in[i]);
    }

  if (SDTR_in[4] < 0 || SDTR_in[4] > 16)	/* if offset out of range */
     {
     printf("\n  ERROR!  Synchronous Offset out of range");
     scsi_error = true;
     return;
     }
  parms.sync_offset = SDTR_in[4];		/* else set sync offset */

  temp1 = SDTR_in[3] * 4;		/* convert period to nanoseconds */

  if (parms.FASTclk == 'n')			/* if 25 MHz clock */
    {
    if (temp1 < 200)				/* if period out of range */
      {
      printf("\nERROR! Synchronous Period too small");
      printf("\n  200ns is the smallest permisible period with a 25MHz clock");
      printf("\n  period = %d ns",temp1);
      scsi_error = true;
      return;
      }
    parms.sync_period = (SDTR_in[3] / 10) & 0x1f;	/* else set SDTR */
    }

  else						/* 40 MHz clock */
    {
    if (temp1 < 100)				/* if period out of range */
      {
      printf("\nERROR! Synchronous Period too small");
      printf("\n  100ns minimum with a 40MHz clock");
      printf("\n  period = %d ns",temp1);
      scsi_error = true;
      return;
      }

    if (temp1 < 200) parms.FAST_SCSI = 'y';
                       
    temp1 = temp1 * 10;			/* gyrations for integer rounding */
    temp1 = temp1 / 25;
    temp1 = temp1 + 5;
    temp1 = temp1 / 10;

    parms.sync_period = temp1 & 0x1f;
    }	/* end of if(FAST_SCSI) */

  load_regs();					/* load registers */

  if(parms.detail=='y')
    {
    printf ("\n  Period: %dnS   Offset: %d",(SDTR_in[3] * 4),SDTR_in[4]);
    printf ("\n  Synchronous Transfer Period Register set to: %d",parms.sync_period);
    printf ("\n  Synchronous Offset Register set to: %d\n",parms.sync_offset);
    }

  if ((inportb(stat_reg) & 7) == 2)	/* Command Phase ? */
    {
    if(parms.detail=='y') printf("\n\nTest Unit Ready\n\nCommand Out ... ");
    outportb(cmd_reg, clr_fifo);
    outportb(cmd_reg,0x98);		/* hackers Test Unit Ready Command */
    isr(19);
    if (scsi_error) return;

    cmdcomp();				/* command complete sequence */
    outportb(cmd_reg,0x80);		/* clear Transfer Count Zero bit */
    isr(3);				/* expect Status and Message good */
    if (scsi_error) return;

    message();				/* issue message accepted cmd */
    isr(4);				/* expect Disconnected Interrupt */
    }
  }	/* end of valid SDTR message */
else if (SDTR_in[0] == 0x07)
  {
  if (parms.detail == 'y') printf("Message Reject");
  scsi_error = true;
  }
}	/* end of synch_nego() */

/************************** Enhanced SRB Mode Test ***********************

This module tests NCR's new Enhanced Save Residual Byte Mode in a
scatter/gather environment.  In this test, three blocks of data are read
in from the Target and "scattered" around memory in three seperate
locations -- all of which are "misaligned" (start on an odd address).
The read occurs for synchronous Data In, which is the most difficult case
to handle.  This particular test uses a single contigious block of memory,
which happens to be word aligned, but the test breaks it up into three
blocks of 512 bytes and forces each block to begin at an odd address.

Block 1:
 - 512 bytes
 - begins at address offset 1
 - ends at   adress offset 512

Block 2:
 - 512 bytes
 - begins at address offset 513
 - ends at   address offset 1024

Block 3:
 - 512 bytes
 - begins at address offset 1025
 - ends at   address offset 1535 + 1 byte wraparound
   the last wraparound byte is stuffed back into element zero of the array.

The first byte of each block (address 1, 513 and 1025) is transfered using
the new ESRB mode.  This requires a DMA Transfer with the byte counter
set to one.  The next 510 bytes in each block are word aligned and are
transfered using normal 16-bit DMA.  Now, each block has been filled with
a total of 511 bytes (through addresss 511, 1023 and 1535).  The last byte
of each block (address 512, 1024 and 1536) is removed using the old SRB
method, where the transfer counter is loaded with an odd number (511)
which causes the chips to end the DMA transfer when the count drops below
2, so the last byte is simply read out of the FIFO.

                             ***************/

void esrb(void)
{
pause_flag = true;
test_error = false;

if(parms.detail == 'y') printf("\nMisaligned scatter/gather test");

if (parms.blkcount != 3)
  {
  printf("\n  ERROR! block count = %d; must be 3.",parms.blkcount);
  test_error = true;
  }

if (RCB != 0x11 && RCB != 0xA2)
  {
  printf("\n  ERROR! wrong version of chip.  RCB = %02X,",RCB);
  printf("\n         ESRB only exists in versions 11 and A2");
  return;
  }

if (parms.blksize != 512)
  {
  printf("\n  ERROR! block size = %d; must be 512.",parms.blksize);
  test_error = true;
  }
if (parms.sync_offset == 0)
  {
  printf("\n  ERROR! synchronous offset = 0; must be synchronous (1-15)");
  test_error = 2;
  }
if (test_error)
  {
  printf("\n  make corrections in (s)etup menu");
  if (test_error == 2)
    {
    printf(", then (e)xecute a SCSI command\n  for Synchronous Negotiation\n");
    }
  return;
  }

/* begin test here ... Synchronous Data In for misaligned, scattered block */

cdb[0]=0x08;				/* SCSI read              */
cdb[1]=0x00;				/* lun = 0, msbaddr = 0	  */
cdb[2]=0x01;				/* msb address = 01       */
cdb[3]=0x00;				/* lsb address = 00	  */
cdb[4]=parms.blkcount;			/* number blocks to xfer  */
cdb[5]=0x00;				/* reserved byte	  */

select(6);				/* select a device */
isr(1);					/* expect Data In */
if(scsi_error) return;			/* exit, abort */
while ((status & 7) == 7)		/* Message In ? */
  {
  Message_In('r');
  if (scsi_error) return;
  }

outportb(config4, (config4_image | 1));		/* Enable ESRB */
outportb(config3, (config3_image | 4));		/* Enhanced SRB */

						/* get byte #1 */
if(parms.detail == 'y') printf("\nData In, byte 1 ......... ");
outportb(tc_high, 0);
outportb(tc_low, 1);
outportb(cmd_reg, DMA_Transfer_Info);
isr(23);
if(scsi_error) return;
dmabyte[1] = inportb(fifo);
						/* get bytes 2-511 */
if(parms.detail == 'y') printf("\nData In, byte 2-511 ..... ");
outportb(tc_high, 0x01);
outportb(tc_low, 0xff);
initIBMdma(&dmabyte[2], 510, 'r');
outportb(cmd_reg, DMA_Transfer_Info);
isr(23);
if (scsi_error) return;
						/* get byte #512 */
if(parms.detail == 'y') printf("\nData In, byte 512 ....... ");
dmabyte[512] = inportb(fifo);
if(parms.detail == 'y') printf("complete");

						/* get byte #513 */
if(parms.detail == 'y') printf("\nData In, byte 513 ....... ");
outportb(tc_high, 0);
outportb(tc_low, 1);
outportb(cmd_reg, DMA_Transfer_Info);
isr(23);
if(scsi_error) return;
dmabyte[513] = inportb(fifo);

						/* get bytes 514-1023 */
if(parms.detail == 'y') printf("\nData In, byte 514-1023 ... ");
outportb(tc_high, 0x01);
outportb(tc_low, 0xff);
initIBMdma(&dmabyte[514], 510, 'r');
outportb(cmd_reg, DMA_Transfer_Info);
isr(23);
if (scsi_error) return;

						/* get byte 1024 */
if(parms.detail == 'y') printf("\nData In, byte 1024 ....... ");
dmabyte[1024] = inportb(fifo);
if(parms.detail == 'y') printf("complete");

						/* get byte 1025 */
if(parms.detail == 'y') printf("\nData In, byte 1025 ....... ");
outportb(tc_high, 0);
outportb(tc_low, 1);
outportb(cmd_reg, DMA_Transfer_Info);
isr(23);
if(scsi_error) return;
dmabyte[1025] = inportb(fifo);

						/* get bytes 1026 - 1535 */
if(parms.detail == 'y') printf("\nData In, byte 1025-1534 ... ");
outportb(tc_high, 0x01);
outportb(tc_low, 0x0ff);
initIBMdma(&dmabyte[1026], 510, 'r');
outportb(cmd_reg, DMA_Transfer_Info);
isr(24);
if (scsi_error) return;

						/* get byte 1536 */
if(parms.detail == 'y') printf("\nData In; byte 1536 ........ ");
dmabyte[0] = inportb(fifo);		/* put last byte in element zero */
if(parms.detail == 'y') printf("complete");

cmdcomp();				/* command complete sequence */
isr(3);					/* expect Status and Message good */
if(scsi_error) return;			/* exit, abort */

message();				/* issue message accepted cmd */
isr(4);					/* expect Disconnected Interrupt */

}

/*************************** Read 53C406 Registers **********************

This routine reads the register in the 53C406

                                  ***********/

void read_regs(void)
{
clrscr();
printf("\n\nThe current 53C94 register contents are: \n");

printf("\n00  Transfer Counter                 =   %02X",inportb(tc_low));
printf("\n01  Transfer Counter                 =   %02X",inportb(tc_high));
printf("\n02  FIFO                             =   %02X",inportb(fifo));
printf("\n03  Command Register                 =   %02X",inportb(cmd_reg));
printf("\n04  Status Register                  =   %02X",inportb(stat_reg));
printf("\n04  Interrupt Status Register        =   %02X",inportb(int_reg));
printf("\n06  Sequence Step Register           =   %02X",inportb(seq_reg));
printf("\n07  FIFO Flags                       =   %02X",inportb(flags_reg));
printf("\n08  Configuration 1 Register         =   %02X",inportb(config1));
printf("\n09  Clock Conversion Factor              write only");
printf("\n0A  Test Register                        write only");
printf("\n0B  Configuration 2 Register         =   %02X",inportb(config2));
printf("\n0C  Configuration 3 Register         =   %02X",inportb(config3));
printf("\n0D  Configuration 4 Register         =   %02X",inportb(config4));
printf("\n0E  TC High/ID Register              =   %02X",inportb(e_reg));
printf("\n0F  FIFO Bottom Register                 write only");

printf("\n\nHit any key to return to menu");
temp = getche();

} /* end of read_regs() */

